import sys
import os
import re
import argparse
import glob
import subprocess
import tempfile

import qbfolkman
import qdimacs
import qcir

def test_generation_encoding(output, maxorder):
    maxseqlength = maxorder - 3

    print("new_int(m, 2, %d)" % (maxorder  - 2), file=output)
    print("new_int(k, 2, %d)" % (maxorder - 2 + maxseqlength - 1), file=output)
    print("new_int(l, 2, %d)" % maxseqlength, file=output)
    print("new_int(ar, 2, %d)" % (maxorder - 2), file=output)

    for i in range(0, maxseqlength - 1):
        print("new_int(a%d, 0, %d)" % (i + 1, maxorder - 2), file=output)
        print("new_bool(x%d)" % (i + 1), file=output)
        print("int_leq(a%d, ar)" % (i + 1), file=output)
        print("int_neq(a%d, 1)" % (i + 1), file=output)
        print("int_neq_reif(a%d, 0, x%d)" % (i + 1, i + 1), file=output)

    print("int_array_sum_eq([%s, ar], k)" % ",".join(["a%d" % (i + 1) for i in range(0, maxseqlength - 1)]), file=output)
    print("int_array_sumCond_eq([%s, 1], [%s, 1], l)" % (",".join(["x%d" % (i + 1) for i in range(0, maxseqlength - 1)]), ",".join(["1" for i in range(0, maxseqlength - 1)])), file=output)
    print("int_array_sum_eq([m, l, -1], k)", file=output)
    print("int_array_sum_leq([ar, m], %d)" % maxorder, file=output)
    print("int_arrays_lex([%s], [%s,ar])" % (",".join(["a%d" % (i + 1) for i in range(0, maxseqlength -1)]), ",".join(["a%d" % (i + 1) for i in range(1, maxseqlength - 1)])), file=output)
    print("solve satisfy(0)", file=output)

def run(testfname, solver, solver_args, timeout, repeat):
    args = ["timeout", str(timeout)] if timeout is not None else []
    args += ["perf", "stat", "-C", "0", "-x,"]

    if repeat is not None:
        args += ["-r", str(repeat)]

    args += ["cgexec", "-g", "cpuset,memory:qbfolkman", solver]
    args += solver_args
    args += [testfname]

    with subprocess.Popen(args, stdout=subprocess.DEVNULL, stderr=subprocess.PIPE, universal_newlines=True) as solverrun:
        res = solverrun.wait()

        if res == 124:
            return None
        else:
            assert(res == 10)
            cpuclockline = ""

            while "cpu-clock" not in cpuclockline:
                cpuclockline = solverrun.stderr.readline().strip()

            runinfo = cpuclockline.split(",")

            return float(runinfo[0])

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Generate experiments for the qbfolkman project")

    parser.add_argument("--PG", help="Use the Plaisted-Greenbaum encoding", default=False, action="store_true")
    parser.add_argument("--isolators", "-i", help="Directory for isolators")
    parser.add_argument("--nauty", "-n", help="Directory for nauty (to use in partial isolators for QDIMACS)")
    parser.add_argument("--cardiso", "-c", help="Use cardinality-based partial isolators for QCIR", default=False, action="store_true")
    parser.add_argument("--label", "-l", help="Label for the test", default="test")
    parser.add_argument("--timeout", "-t", type=int, help="Timeout in seconds")
    parser.add_argument("--repeat", "-r", type=int, help="Repeat run N times", metavar="N")
    parser.add_argument("--solver-arg", "-s", help="Solver arguments", action="append", default=[])
    parser.add_argument("--latex", "-L", help="Use LaTeX output", default=False, action="store_true")
    parser.add_argument("--bloqqer", "-b", help="Preprocess QDIMACS formulas with bloqqer")
    parser.add_argument("--tempdir", "-d", help="Temporary directory for files")

    parser.add_argument("format", help="QDIMACS or QCIR", choices=["qdimacs", "qcir"])
    parser.add_argument("bee", help="Location of the BEE directory")
    parser.add_argument("m", type=int, help="Max order of graphs for test")
    parser.add_argument("solver", help="Solver to run for tests (must accept the selected format)")

    res = parser.parse_args()
    are = re.compile("a[^\s+] = (\d+)")
    a = []
    preva = set()
    isolators = {}

    if res.isolators:
        for i in glob.glob(os.path.join(res.isolators, "*.cnf")):
            with open(i, 'r') as isofile:
                iso, nvars = qbfolkman.read_dimacs(isofile)
                order = int(os.path.basename(i)[:-len(".cnf")])
                isolators[order] = i

    with tempfile.NamedTemporaryFile(mode="w", dir=res.tempdir) as encoding:
        test_generation_encoding(encoding, res.m)
        encoding.flush()

        with subprocess.Popen(["./BumbleBEE", encoding.name], cwd=res.bee, stdout=subprocess.PIPE, universal_newlines=True) as BumbleBEE:
            for l in BumbleBEE.stdout:
                l = l.strip()

                if l == "----------":
                    if tuple(sorted(a)) not in preva:
                        preva.add(tuple(sorted(a)))

                    a = []
                else:
                    m = are.match(l)

                    if m is not None:
                        aval = int(m.group(1))

                        if aval > 0:
                            a.append(aval)

    print("About to test %d cases" % len(preva), file=sys.stderr)

    for a in sorted(preva, key=lambda x: (sum([i - 1 for i in x]) + max(x), len(x))):
        isolator = None

        if res.isolators:
            m, ar = qbfolkman.parameters(a)

            if (m + ar) in isolators:
                isolator = isolators[m + ar]
            else:
                print("WARNING: no isolator for", a, file=sys.stderr)


        if res.cardiso:
            tmpfile, tmpname = tempfile.mkstemp(dir=res.tempdir)

            subprocess.check_output(["../build/qcir", "-d", tmpname] + [str(i) for i in a])

            runmsec = run(tmpname, res.solver, res.solver_arg, res.timeout, res.repeat)

            if runmsec is not None:
                print(("(%s)\t%f" if not res.latex else "\\texttt{%s}&%.2f") % (",".join([str(i) for i in a]), runmsec))
            else:
                print(("(%s)\t-" if not res.latex else "\\texttt{%s}&-") % ",".join([str(i) for i in a]))

            os.unlink(tmpname)
        else:
            with tempfile.NamedTemporaryFile(mode="w", dir=res.tempdir) as outfile:
                if res.format == "qdimacs":
                    if isolator is not None:
                        with open(isolator, 'r') as isofile:
                            qdimacs.qdimacs(a, outfile, res.PG, isolator=isofile, nautydir=res.nauty)
                    else:
                        if res.bloqqer is not None:
                            with subprocess.Popen([res.bloqqer], stdin=subprocess.PIPE, stdout=outfile, universal_newlines=True) as bloqqer:
                                qdimacs.qdimacs(a, bloqqer.stdin, res.PG, nautydir=res.nauty)
                        else:
                            qdimacs.qdimacs(a, outfile, res.PG, nautydir=res.nauty)
                elif res.format == "qcir":
                    if isolator is not None:
                        with open(isolator, 'r') as isofile:
                            qcir.qcir(a, outfile, isolator=isofile)

                outfile.flush()
                runmsec = run(outfile.name, res.solver, res.solver_arg, res.timeout, res.repeat)

                if runmsec is not None:
                    print(("(%s)\t%f" if not res.latex else "\\texttt{%s}&%.2f") % (",".join([str(i) for i in a]), runmsec))
                else:
                    print(("(%s)\t-" if not res.latex else "\\texttt{%s}&-") % ",".join([str(i) for i in a]))
