#include "circuits.h"

#include <assert.h>

#include <gsl/gsl_combination.h>

#include <solver.h>
#include <qcir.h>

#include <nauty.h>

/**
 * These variables are global because of the terrible geng API
 */
Circuit * formula;
Gate * canon_or;
size_t N = 0;
Gate * edges_pinned = NULL;

var_t edge_var(size_t v, size_t u, size_t n)
{
  int i = (v > u) ? v : u;
  int j = (v > u) ? u : v;

  return i * (i - 1) / 2 + j + 1;
}

var_t vtx_color_var(size_t v, size_t c, size_t n, int maxid)
{
  return maxid + c * n + v + 1;
}

void geng_out(FILE *f, graph *g, int n)
{
  Gate * and = circuit_add_gate(formula, formula->max_num + 1, GATE_AND);

  for (int i = 0; i < n; i++) {
    for (int j = i + 1; j < n; j++) {
      if (ISELEMENT(GRAPHROW(g, i, 1), j)) {
        circuit_add_to_gate(formula, and, create_lit(edge_var(i, j, (N > 0) ? N : n), (N > 0)));
      }
      else {
        circuit_add_to_gate(formula, and, create_lit(edge_var(i, j, (N > 0) ? N : n), (N == 0)));
      }
    }
  }

  if (edges_pinned) {
    circuit_add_to_gate(formula, and, create_lit(edges_pinned->shared.orig_id, false));
  }

  circuit_add_to_gate(formula, canon_or, create_lit(and->shared.orig_id, false));
}

var_t graph_search(size_t n, size_t m, Circuit * formula)
{
  gsl_combination * cliques = gsl_combination_calloc(n, m);
  Gate * and = circuit_add_gate(formula, formula->max_num + 1, GATE_AND);

  do {
    Gate * or = circuit_add_gate(formula, formula->max_num + 1, GATE_OR);
    gsl_combination * edges = gsl_combination_calloc(m, 2);

    do {
      size_t v = gsl_combination_get(cliques, gsl_combination_get(edges, 0));
      size_t u = gsl_combination_get(cliques, gsl_combination_get(edges, 1));

      circuit_add_to_gate(formula, or, create_lit(edge_var(u, v, n), true));
    } while (gsl_combination_next(edges) == GSL_SUCCESS);

    gsl_combination_free(edges);
    circuit_add_to_gate(formula, and, or->shared.orig_id);
  } while (gsl_combination_next(cliques) == GSL_SUCCESS);

  gsl_combination_free(cliques);

  return and->shared.orig_id;
}

var_t unique_coloring(size_t n, size_t c, int maxid, Circuit * formula)
{
  Gate * and = circuit_add_gate(formula, formula->max_num + 1, GATE_AND);

  for (size_t i = 0; i < n; i++) {
    Gate * at_least_or = circuit_add_gate(formula, formula->max_num + 1, GATE_OR);
    gsl_combination * pairs = gsl_combination_calloc(c, 2);

    for (size_t j = 0; j < c; j++) {
      circuit_add_to_gate(formula, at_least_or, create_lit(vtx_color_var(i, j, n, maxid), false));
    }

    do {
      Gate * at_most_or = circuit_add_gate(formula, formula->max_num + 1, GATE_OR);
      var_t color0 = vtx_color_var(i, gsl_combination_get(pairs, 0), n, maxid);
      var_t color1 = vtx_color_var(i, gsl_combination_get(pairs, 1), n, maxid);

      circuit_add_to_gate(formula, at_most_or, create_lit(color0, true));
      circuit_add_to_gate(formula, at_most_or, create_lit(color1, true));
      circuit_add_to_gate(formula, and, at_most_or->shared.orig_id);
    } while (gsl_combination_next(pairs) == GSL_SUCCESS);

    gsl_combination_free(pairs);
    circuit_add_to_gate(formula, and, at_least_or->shared.orig_id);
  }

  return and->shared.orig_id;
}

var_t coloring_search(size_t n, size_t * a, size_t colors, int maxid, Circuit * formula)
{
  Gate * or = circuit_add_gate(formula, formula->max_num + 1, GATE_OR);

  for (size_t i = 0; i < colors; i++) {
    gsl_combination * cliques = gsl_combination_calloc(n, a[i]);

    do {
      Gate * and = circuit_add_gate(formula, formula->max_num + 1, GATE_AND);
      gsl_combination * edges = gsl_combination_calloc(a[i], 2);

      do {
        size_t v = gsl_combination_get(cliques, gsl_combination_get(edges, 0));
        size_t u = gsl_combination_get(cliques, gsl_combination_get(edges, 1));

        circuit_add_to_gate(formula, and, create_lit(edge_var(u, v, n), false));
      } while (gsl_combination_next(edges) == GSL_SUCCESS);

      for (size_t j = 0; j < a[i]; j++) {
        var_t vc = vtx_color_var(gsl_combination_get(cliques, j), i, n, maxid);

        circuit_add_to_gate(formula, and, create_lit(vc, false));
      }

      gsl_combination_free(edges);
      circuit_add_to_gate(formula, or, and->shared.orig_id);
    } while (gsl_combination_next(cliques) == GSL_SUCCESS);

    gsl_combination_free(cliques);
  }

  return or->shared.orig_id;
}

int main(int argc, char ** argv)
{
  size_t ar = 0;
  size_t m = 1;
  char * debug = NULL;
  bool pin = false;
  int optc;

  while ((optc = getopt(argc, argv, "d:p")) != -1) {
    switch (optc) {
    case 'd':
      asprintf(&debug, "%s", optarg);
      break;
    case 'p':
      pin = true;
      break;
    default:
      fprintf(stderr, "Unknown argument %c\nUsage: %s [-d] [-p]\n", optc, argv[0]);
      break;
    }
  }

  size_t colors = argc - optind;
  size_t * a = malloc(colors * sizeof(size_t));

  for (int i = optind; i < argc; i++) {
    if (atol(argv[i]) < 1) {
      free(a);

      return -1;
    }

    a[i - optind] = atol(argv[i]);

    if (a[i - optind] > ar) {
      ar = a[i - optind];
    }
  }

  if (ar == 0) {
    free(a);

    return -1;
  }

  for (size_t i = 0; i < colors; i++) {
    m += a[i] - 1;
  }

  SolverOptions * opt = solver_get_default_options();
  Scope * exists, * forall;
  int maxedgeid = edge_var(ar + m - 2, ar + m - 1, ar + m);

  formula = circuit_init();
  exists = circuit_init_scope(formula, QUANT_EXISTS);
  forall = circuit_init_scope(formula, QUANT_FORALL);

  for (size_t i = 0; i < ar + m; i++) {
    for (size_t j = i + 1; j < ar + m; j++) {
      circuit_new_var(formula, exists, edge_var(i, j, ar + m));
    }
  }

  for (size_t i = 0; i < ar + m; i++) {
    for (size_t j = 0; j < colors; j++) {
      circuit_new_var(formula, forall, vtx_color_var(i, j, ar + m, maxedgeid));
    }
  }

  var_t g = graph_search(ar + m, m, formula);
  var_t u = unique_coloring(ar + m, colors, maxedgeid, formula);
  var_t c = coloring_search(ar + m, a, colors, maxedgeid, formula);
  Gate * or = circuit_add_gate(formula, formula->max_num + 1, GATE_OR);
  Gate * and = circuit_add_gate(formula, formula->max_num + 1, GATE_AND);

  circuit_add_to_gate(formula, or, create_lit(u, true));
  circuit_add_to_gate(formula, or, create_lit(c, false));
  circuit_add_to_gate(formula, and, create_lit(g, false));
  circuit_add_to_gate(formula, and, create_lit(or->shared.orig_id, false));
  circuit_set_output(formula, create_lit(and->shared.orig_id, false));

  if (true) {
    char * geng_argv[3];
    char * geng_name = "geng";
    GArray * neg_edge_lits = g_array_sized_new(FALSE, FALSE, sizeof(lit_t), maxedgeid);
    Gate * isolator = circuit_add_gate(formula, formula->max_num + 1, GATE_OR);
    var_t card_indicator;

    for (size_t i = 0; i < ar + m; i++) {
      for (size_t j = i + 1; j < ar + m; j++) {
        lit_t l = create_lit(edge_var(i, j, ar + m), true);

        g_array_append_val(neg_edge_lits, l);
      }
    }

    card_indicator = card_network(formula, neg_edge_lits, 2 * ar + 1);
    canon_or = circuit_add_gate(formula, formula->max_num + 1, GATE_OR);
    circuit_add_to_gate(formula, isolator, create_lit(card_indicator, false));
    circuit_add_to_gate(formula, isolator, create_lit(canon_or->shared.orig_id, false));
    circuit_add_to_gate(formula, and, create_lit(isolator->shared.orig_id, false));
    geng_argv[0] = geng_name;

    if (pin && ((4 * ar + 2) < (ar + m))) {
      edges_pinned = circuit_add_gate(formula, formula->max_num + 1, GATE_AND);

      for (size_t i = 0; i < ar + m; i++) {
        for (size_t j = i + 1; j < ar + m; j ++) {
          if ((i > (4 * ar + 1)) || (j > (4 * ar + 1))) {
            circuit_add_to_gate(formula, edges_pinned, create_lit(edge_var(i, j, ar + m), false));
          }
        }
      }

      N = ar + m;
      asprintf(&geng_argv[1], "%lu", 4 * ar + 2);
      asprintf(&geng_argv[2], "0:%lu", 2 * ar + 1);
    }
    else {
      asprintf(&geng_argv[1], "%lu", ar + m);
      asprintf(&geng_argv[2], "%lu:%d", maxedgeid - 2 * ar - 1, maxedgeid);
    }

    geng_main(3, geng_argv);
    g_array_unref(neg_edge_lits);
    free(geng_argv[1]);
    free(geng_argv[2]);
  }

  if (debug) {
    circuit_reencode(formula);
    circuit_open_and_write_qcir_file(formula, debug);
    free(debug);
  }
  else {
    Solver * quabs = solver_init(opt, formula);

    assert(solver_sat(quabs) == QBF_RESULT_SAT);
    solver_free(quabs);
  }

  free(opt);
  circuit_free(formula);
  free(a);
}
