import os
import sys
import itertools
import subprocess
import argparse
import re

import igraph

import qbfolkman

import pysat.formula
from pysat.card import CardEnc, EncType

def cnf2qcir(formula, output, lastvarid):
    for i in range(0, len(formula)):
        print("%d = or(%s)" % (i + 1 + lastvarid, ",".join(str(l) for l in formula[i])), file=output)

    print("%d = and(%s)" % (lastvarid + len(formula) + 1, ",".join(["%d" % (i + 1 + lastvarid) for i in range(0, len(formula))])), file=output)

    return lastvarid + len(formula) + 1

def dnf2qcir(formula, output, lastvarid):
    for i in range(0, len(formula)):
        print("%d = and(%s)" % (i + 1 + lastvarid, ",".join(str(l) for l in formula[i])), file=output)

    print("%d = or(%s)" % (lastvarid + len(formula) + 1, ",".join(["%d" % (i + 1 + lastvarid) for i in range(0, len(formula))])), file=output)

    return lastvarid + len(formula) + 1

def qcir(a, output, BreakIDbin=None, isolator=None):
    m, ar = qbfolkman.parameters(a)
    vs = [i for i in range(0, ar + m)]
    ecount = qbfolkman.nchoosek(ar + m, 2)
    graphsearchcnf = []
    edgeboundcnf = []
    canondnf = []
    uniquecnf = []
    foralldnf = []
    lastvarid = qbfolkman.nchoosek(ar + m, 2)

    for clique in itertools.combinations(vs, m):
        graphsearchcnf.append([-qbfolkman.edge(e[0], e[1], ar + m) for e in itertools.combinations(clique, 2)])

    if BreakIDbin is not None:
        with subprocess.Popen([BreakIDbin, "-", "-no-relaxed", "-print-only-breakers"], stdout=subprocess.PIPE, stdin=subprocess.PIPE, universal_newlines=True) as BreakID:
            pysat.formula.CNF(from_clauses=graphsearchcnf).to_fp(BreakID.stdin)
            BreakID.stdin.close()

            symbreakers, lastvarid = qbfolkman.read_dimacs(BreakID.stdout, ignore_header=True)

            graphsearchcnf += symbreakers
    else:
        if isolator is not None:
            isocnf, lastvarid = qbfolkman.read_dimacs(isolator)

            graphsearchcnf += isocnf

    for v in vs:
        eqcnf = CardEnc.equals([qbfolkman.vcolor(v, c, ar + m, lastvarid) for c in range(0, len(a))], encoding=EncType.pairwise)

        for c in eqcnf.clauses:
            uniquecnf.append(c)

    for c in range(0, len(a)):
        for clique in itertools.combinations(vs, a[c]):
            foralldnf.append([qbfolkman.edge(e[0], e[1], ar + m) for e in itertools.combinations(clique, 2)] + [qbfolkman.vcolor(v, c, ar + m, lastvarid) for v in clique])

    print("#QCIR-G14", file=output)
    print("exists(%s)" % ",".join(["%d" % (i + 1) for i in range(0, lastvarid)]), file=output)
    print("forall(%s)" % ",".join(["%d" % (lastvarid + i + 1) for i in range(0, len(a) * (ar + m))]), file=output)

    outvar = lastvarid + len(a) * (ar + m) + 1

    print("output(%d)" % outvar, file=output)

    gvar = cnf2qcir(graphsearchcnf, output, outvar)
    lastvarid = gvar
    uvar = cnf2qcir(uniquecnf, output, lastvarid)
    avar = dnf2qcir(foralldnf, output, uvar)
    cvar = avar + 1

    print("%d = or(-%d, %d)" % (cvar, uvar, avar), file=output)
    print("%s = and(%s, %s)" % (outvar, gvar, cvar), file=output)

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Generate QDIMACS instances of unique Folkman witnesses")

    parser.add_argument("--isolator", "-i", type=argparse.FileType('r'), help="Use the given isolator for the graph search")
    parser.add_argument("--BreakID", "-b", help="Break symmetries in the graph search using BreakID")
    parser.add_argument("--solver", "-s", help="Specify a QCIR-enabled solver")
    parser.add_argument("--verify", "-v", help="Verify the solution", default=False, action="store_true")

    parser.add_argument("seq", type=int, nargs="+", help="Input sequence")

    res = parser.parse_args()

    if res.solver is not None:
        with subprocess.Popen([res.solver], stdin=subprocess.PIPE, stdout=subprocess.PIPE if res.verify else None, universal_newlines=True) as solver:
            qcir(res.seq, solver.stdin, res.BreakID)

            if res.verify:
                m, ar = qbfolkman.parameters(res.seq)
                solution = igraph.Graph(m + ar)
                correct = qbfolkman.correct_answer(res.seq)
                soledges = set()
                vs = [i for i in range(0, ar + m)]
                ere = re.compile("\+(\d+)")

                solver.stdin.close()
                assert(solver.wait() == 10)

                for l in solver.stdout:
                    if l.startswith("v"):
                        for ematch in ere.finditer(l):
                            soledges.add(int(ematch.group(1)))

                for i in range(0, ar + m):
                    for j in range(i + 1, ar + m):
                        if qbfolkman.edge(i, j, ar + m) in soledges:
                            solution.add_edge(i, j)

                assert(solution.isomorphic_bliss(correct))
    else:
        qcir(res.seq, sys.stdout, res.BreakID)
