#include <assert.h>
#include <getopt.h>

#include <gmodule.h>

#include <gsl/gsl_combination.h>

#include <igraph.h>

#include <qdpll.h>

static struct option long_options[] = {
  {"PG", no_argument, 0, 0},
  {0, 0, 0, 0}
};

int edge_var(size_t v, size_t u, size_t n)
{
  int i = (v > u) ? v : u;
  int j = (v > u) ? u : v;

  return i * (i - 1) / 2 + j + 1;
}

int vtx_color_var(size_t v, size_t c, size_t n, int maxid)
{
  return maxid + c * n + v + 1;
}

// Pairwise encoding of ExactlyOneOf
int exactly_one_of(GArray * vars, GSList ** formula)
{
  gsl_combination * pairs = gsl_combination_calloc(vars->len, 2);

  *formula = g_slist_prepend(*formula, g_array_ref(vars));

  do {
    int v1 = -g_array_index(vars, int, gsl_combination_get(pairs, 0));
    int v2 = -g_array_index(vars, int, gsl_combination_get(pairs, 1));
    GArray * pair = g_array_sized_new(FALSE, FALSE, sizeof(int), 2);

    g_array_append_val(pair, v1);
    g_array_append_val(pair, v2);
    *formula = g_slist_prepend(*formula, pair);
  } while (gsl_combination_next(pairs) == GSL_SUCCESS);

  gsl_combination_free(pairs);

  // Ignored for now, but may be useful for other encodings
  return 0;
}

int graph_search_cnf(size_t n, size_t m, GSList ** formula)
{
  gsl_combination * cliques = gsl_combination_calloc(n, m);

  do {
    gsl_combination * edges = gsl_combination_calloc(m, 2);
    GArray * clause = g_array_sized_new(FALSE, FALSE, sizeof(int), m);

    do {
      size_t v = gsl_combination_get(cliques, gsl_combination_get(edges, 0));
      size_t u = gsl_combination_get(cliques, gsl_combination_get(edges, 1));
      int lit = -edge_var(v, u, n);

      g_array_append_val(clause, lit);
    } while (gsl_combination_next(edges) == GSL_SUCCESS);

    gsl_combination_free(edges);
    *formula = g_slist_prepend(*formula, clause);
  } while (gsl_combination_next(cliques) == GSL_SUCCESS);

  gsl_combination_free(cliques);

  return edge_var(n - 2, n - 1, n);
}

int graph_avoid_dnf(size_t n, size_t colors, size_t * a, GSList ** formula, int maxid)
{
  for (size_t i = 0; i < colors; i++) {
    gsl_combination * cliques = gsl_combination_calloc(n, a[i]);

    do {
      GArray * clause = g_array_sized_new(FALSE, FALSE, sizeof(int), a[i] * (a[i] + 1) / 2);
      gsl_combination * edges = gsl_combination_calloc(a[i], 2);

      do {
	int v = gsl_combination_get(cliques, gsl_combination_get(edges, 0));
	int u = gsl_combination_get(cliques, gsl_combination_get(edges, 1));
	int var = edge_var(v, u, n);

	g_array_append_val(clause, var);
      } while (gsl_combination_next(edges) == GSL_SUCCESS);

      for (size_t j = 0; j < a[i]; j++) {
	int var = vtx_color_var(gsl_combination_get(cliques, j), i, n, maxid);

	g_array_append_val(clause, var);
      }

      *formula = g_slist_append(*formula, clause);
      gsl_combination_free(edges);
    } while (gsl_combination_next(cliques) == GSL_SUCCESS);

    gsl_combination_free(cliques);
  }
}

int dnf_to_cnf(GSList * formula_dnf, GSList ** formula, int maxid, gboolean use_PG)
{
  int aux = maxid;
  GArray * last_clause = g_array_sized_new(FALSE, FALSE, sizeof(int), g_slist_length(formula_dnf));

  while (formula_dnf) {
    GArray * dnf_clause = (GArray *)formula_dnf->data;
    GArray * cnf_clause = NULL;
    int negaux = -(++aux);

    if (!use_PG) {
      cnf_clause = g_array_sized_new(FALSE, FALSE, sizeof(int), dnf_clause->len + 1);
      g_array_append_val(cnf_clause, aux);
    }

    for (size_t i = 0; i < dnf_clause->len; i++) {
      GArray * binary_clause = g_array_sized_new(FALSE, FALSE, sizeof(int), 2);
      int dnfvar = g_array_index(dnf_clause, int, i);
      int negdnfvar = -dnfvar;

      g_array_append_val(binary_clause, negaux);
      g_array_append_val(binary_clause, dnfvar);
      *formula = g_slist_prepend(*formula, binary_clause);

      if (!use_PG) {
	g_array_append_val(cnf_clause, negdnfvar);
      }
    }

    if (!use_PG) {
      *formula  = g_slist_prepend(*formula, cnf_clause);
    }

    g_array_append_val(last_clause, aux);
    formula_dnf = formula_dnf->next;
  }

  *formula = g_slist_prepend(*formula, last_clause);

  return aux;
}

void add_clause(gpointer data, gpointer depqbf)
{
  GArray * clause = (GArray *)data;

  for (guint i = 0; i < clause->len; i++) {
    qdpll_add((QDPLL *)depqbf, g_array_index(clause, int, i));
  }

  qdpll_add((QDPLL *)depqbf, 0);
}

void search_for_graphs(GSList * formula, size_t colors, size_t * a, size_t ar, size_t m, int maxid, int auxmaxid, gboolean verify)
{
  QDPLL * depqbf = qdpll_create();
  GSList * graphs = NULL;

  qdpll_configure (depqbf, "--dep-man=simple");
  qdpll_configure (depqbf, "--incremental-use");

  // Graph search variables
  qdpll_new_scope_at_nesting(depqbf, QDPLL_QTYPE_EXISTS, 1);

  for (int i = 0; i < maxid; i++) {
    qdpll_add(depqbf, i + 1);
  }

  qdpll_add(depqbf, 0);

  // Graph coloring variables
  qdpll_new_scope_at_nesting(depqbf, QDPLL_QTYPE_FORALL, 2);

  for (size_t i = 0; i < colors; i++) {
    for (size_t j = 0; j < (ar + m); j++) {
      qdpll_add(depqbf, vtx_color_var(j, i, ar + m, maxid));
    }
  }

  qdpll_add(depqbf, 0);

  // Auxiliary variables
  qdpll_new_scope_at_nesting(depqbf, QDPLL_QTYPE_EXISTS, 3);

  for (int i = maxid +  colors * (ar + m); i < auxmaxid; i++) {
    qdpll_add(depqbf, i + 1);
  }

  qdpll_add(depqbf, 0);

  g_slist_foreach(formula, add_clause, depqbf);

  while (qdpll_sat(depqbf) == QDPLL_RESULT_SAT) {
    GArray * blocking_clause = g_array_sized_new(FALSE, FALSE, sizeof(int), (ar + m) * (ar + m - 1) / 2);
    igraph_t * graph;

    if (verify) {
      graph = malloc(sizeof(igraph_t));
      igraph_empty(graph, ar + m, FALSE);
    }

    for (size_t i = 0; i < (ar + m); i++) {
      for (size_t j = i + 1; j < (ar + m); j++) {
	int var = edge_var(i, j, ar + m);
	int neglit = -var * qdpll_get_value(depqbf, var);

	assert(qdpll_get_value(depqbf, var) != QDPLL_ASSIGNMENT_UNDEF);

	if ((qdpll_get_value(depqbf, var) == QDPLL_ASSIGNMENT_TRUE) && verify) {
	  igraph_add_edge(graph, i, j);
	}

	g_array_append_val(blocking_clause, neglit);
      }
    }

    GSList * graph_iter = graphs;
    igraph_bool_t is_isomorphic = FALSE;

    if (verify) {
      while (graph_iter && !is_isomorphic) {
	igraph_isomorphic_bliss(graph, (igraph_t *)graph_iter->data, NULL, NULL, &is_isomorphic, NULL, NULL, IGRAPH_BLISS_F, NULL, NULL);
	graph_iter = graph_iter->next;
      }

      if (!is_isomorphic) {
	graphs = g_slist_prepend(graphs, graph);
      }
      else {
	igraph_destroy(graph);
	free(graph);
      }
    }

    qdpll_reset(depqbf);

    for (size_t i = 0; i < blocking_clause->len; i++) {
      qdpll_add(depqbf, g_array_index(blocking_clause, int, i));
    }

    qdpll_add(depqbf, 0);
    g_array_unref(blocking_clause);
  }

  if (verify) {
    assert(g_slist_length(graphs) == 1);
    igraph_write_graph_edgelist((igraph_t *)graphs->data, stdout);
    igraph_destroy(graphs->data);
    g_slist_free_full(graphs, free);
  }

  qdpll_delete(depqbf);
}

int main(int argc, char ** argv)
{
  size_t ar = 0;
  size_t m = 1;
  GSList * formula = NULL;
  GSList * formula_dnf = NULL;
  gboolean use_PG = FALSE;
  gboolean verify = FALSE;

  while (1) {
    int option_index = 0;
    int c = getopt_long(argc, argv, "v", long_options, &option_index);

    if (c == -1) {
      break;
    }

    switch (c) {
    case 0:
      assert(optarg == 0);
      use_PG = TRUE;
      break;
    case 'v':
      verify = TRUE;
    }
  }

  size_t colors = argc - optind;
  size_t * a = malloc(colors * sizeof(size_t));

  for (int i = optind; i < argc; i++) {
    if (atol(argv[i]) < 1) {
      free(a);

      return -1;
    }

    a[i - optind] = atol(argv[i]);

    if (a[i - optind] > ar) {
      ar = a[i - optind];
    }
  }

  if (ar == 0) {
    free(a);

    return -1;
  }

  for (size_t i = 0; i < colors; i++) {
    m += a[i] - 1;
  }

  printf("c a: ");

  for (size_t i = 0; i < colors; i++) {
    printf("%lu ", a[i]);
  }

  printf("\nc ar: %lu\nc m: %lu\n", ar, m);

  int maxid = graph_search_cnf(ar + m, m, &formula);

  for (size_t i = 0; i < (ar + m); i++) {
    GArray * vars = g_array_sized_new(FALSE, FALSE, sizeof(int), colors);

    for (size_t j = 0; j < colors; j++) {
      int lit = -vtx_color_var(i, j, ar + m, maxid);

      g_array_append_val(vars, lit);
    }

    exactly_one_of(vars, &formula_dnf);
    g_array_unref(vars);
  }

  graph_avoid_dnf(ar + m, colors, a, &formula_dnf, maxid);
  int auxmaxid = dnf_to_cnf(formula_dnf, &formula, maxid +  colors * (ar + m), use_PG);

  search_for_graphs(formula, colors, a, ar, m, maxid, auxmaxid, verify);
  g_slist_free_full(formula_dnf, g_array_unref);
  g_slist_free_full(formula, g_array_unref);

  free(a);

  return 0;
}
