import sys
import os
import itertools
import subprocess
import argparse
import re

import igraph

import qbfolkman

import pysat.formula
from pysat.card import CardEnc, EncType

def dnf2cnf(nvars, formula, use_PG):
    lastclause = []
    res = []

    for c in formula:
        nvars += 1
        impclause = []
        lastclause.append(nvars)

        for l in c:
            if not use_PG:
                impclause.append(-l)

            res.append([-nvars, l])

        if not use_PG:
            impclause.append(nvars)
            res.append(impclause)

    res.append(lastclause)

    return nvars, res

def qdimacs(a, output, use_PG, BreakIDbin=None, blocking_clauses=[], isolator=None, nautydir=None):
    m, ar = qbfolkman.parameters(a)
    vs = [i for i in range(0, ar + m)]
    graphsearchcnf = []
    foralldnf = []
    ecount = qbfolkman.nchoosek(ar + m, 2)
    lastvarid = ecount
    cliquecount = qbfolkman.nchoosek(ar + m, m)
    existential = [i + 1 for i in range(0, ecount)]
    universal = []
    auxiliary = []

    for clique in itertools.combinations(vs, m):
        graphsearchcnf.append([-qbfolkman.edge(e[0], e[1], ar + m) for e in itertools.combinations(clique, 2)])

    if BreakIDbin is not None:
        with subprocess.Popen([BreakIDbin, "-", "-no-relaxed", "-print-only-breakers"], stdout=subprocess.PIPE, stdin=subprocess.PIPE, universal_newlines=True) as BreakID:
            pysat.formula.CNF(from_clauses=graphsearchcnf).to_fp(BreakID.stdin)
            BreakID.stdin.close()

            symbreakers, lastvarid = qbfolkman.read_dimacs(BreakID.stdout, ignore_header=True)

            graphsearchcnf += symbreakers
            existential += [(i + 1) for i in range(ecount, lastvarid)]
    elif isolator is not None:
            isocnf, lastvarid = qbfolkman.read_dimacs(isolator)

            graphsearchcnf += isocnf
            existential += [(i + 1) for i in range(ecount, lastvarid)]
    elif nautydir is not None:
        # Create partial isolator
        dnfaux = []
        edgecount = CardEnc.atmost([-qbfolkman.edge(e[0], e[1], ar + m) for e in itertools.combinations(vs, 2)], bound=2 * ar + 1, top_id=ecount)
        isolatordnf = []

        universal  += [i + 1 for i in range(ecount, edgecount.nv)]

        for c in edgecount.clauses:
            isolatordnf.append([-l for l in c])

        with subprocess.Popen([os.path.join(nautydir, "geng"), str(ar + m), "%d:%d" % (ecount - (2 * ar + 1), ecount)], stdout=subprocess.PIPE, universal_newlines=True) as geng:
            with subprocess.Popen([os.path.join(nautydir, "listg"), "-d", "-l0"], stdin=geng.stdout, stdout=subprocess.PIPE, universal_newlines=True) as listg:
                for l in listg.stdout:
                    l = l.strip()

                    if l.startswith("n="):
                        eset = set()
                    elif l.endswith(";"):
                        l = l.strip(";")
                        (v, uliststr) = tuple(l.split(":"))
                        ulist = uliststr.split(" ")

                        for u in ulist:
                            if u:
                                eset.add(qbfolkman.edge(int(u), int(v), ar + m))
                    elif l.startswith("$$"):
                        lclause = []

                        for e in itertools.combinations(vs, 2):
                            evar = qbfolkman.edge(e[0], e[1], ar + m)

                            if evar in eset:
                                lclause.append(evar)
                            else:
                                lclause.append(-evar)

                        isolatordnf.append(lclause)

        lastvarid, isolatorcnf = dnf2cnf(edgecount.nv, isolatordnf, use_PG)

        auxiliary += [(i + 1) for i in range(edgecount.nv, lastvarid)]

    for v in vs:
        vcolorvars = [qbfolkman.vcolor(v, c, ar + m, lastvarid) for c in range(0, len(a))]
        eqcnf = CardEnc.equals(vcolorvars, encoding=EncType.pairwise)

        for c in eqcnf.clauses:
            foralldnf.append([-l for l in c])

        universal += vcolorvars

    for c in range(0, len(a)):
        for clique in itertools.combinations(vs, a[c]):
            foralldnf.append([qbfolkman.edge(e[0], e[1], ar + m) for e in itertools.combinations(clique, 2)] + [qbfolkman.vcolor(v, c, ar + m, lastvarid) for v in clique])

    nvars, forallcnf = dnf2cnf(lastvarid + len(a) * (ar + m), foralldnf, use_PG)

    auxiliary += [(i + 1) for i in range(max(universal), nvars)]
    graphsearchcnf += blocking_clauses

    print("p cnf %d %d" % (nvars, len(graphsearchcnf) + len(forallcnf)), file=output)
    print("e %s 0" % " ".join([str(i) for i in existential]), file=output)
    print("a %s 0" % " ".join([str(i) for i in universal]), file=output)
    print("e %s 0" % " ".join([str(i) for i in auxiliary]), file=output)

    for c in graphsearchcnf:
        print("%s 0" % " ".join([str(l) for l in c]), file=output)

    for c in forallcnf:
        print("%s 0" % " ".join([str(l) for l in c]), file=output)

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Generate QDIMACS instances of unique Folkman witnesses")

    parser.add_argument("--isolator", "-i", type=argparse.FileType('r'), help="Use the given isolator for the graph search")
    parser.add_argument("--BreakID", "-b", help="Break symmetries in the graph search using BreakID")
    parser.add_argument("--nauty", "-n", help="Break symmetries in the graph search using partial isolators out of nauty")
    parser.add_argument("--PG", help="Use the Plaisted-Greenbaum encoding", default=False, action="store_true")
    parser.add_argument("--depqbf", "-d", help="Specify location of DepQBF")
    parser.add_argument("--verify", "-v", help="Verify the solution", default=False, action="store_true")

    parser.add_argument("seq", type=int, nargs="+", help="Input sequence")

    res = parser.parse_args()

    if res.depqbf is not None:
        with subprocess.Popen([res.depqbf, "--qdo", "--no-dynamic-nenofex"], stdin=subprocess.PIPE, stdout=subprocess.PIPE if res.verify else None, universal_newlines=True) as solver:
            qdimacs(res.seq, solver.stdin, res.PG, res.BreakID, isolator=res.isolator, nautydir=res.nauty)

            if res.verify:
                m, ar = qbfolkman.parameters(res.seq)
                correct = qbfolkman.correct_answer(res.seq)
                solution = igraph.Graph(m + ar)
                soledges = set()
                vs = [i for i in range(0, ar + m)]
                vre = re.compile("V (-?\d+) 0\n")
                blocking = []

                solver.stdin.close()
                assert(solver.wait() == 10)

                for l in solver.stdout:
                    vmatch = vre.match(l)

                    if vmatch is not None:
                        l = int(vmatch.group(1))

                        blocking.append(-l)

                        if l > 0:
                            soledges.add(l)

                for i in range(0, ar + m):
                    for j in range(i + 1, ar + m):
                        if qbfolkman.edge(i, j, ar + m) in soledges:
                            solution.add_edge(i, j)

                assert(solution.isomorphic_bliss(correct))

                if res.isolator is not None:
                    res.isolator.seek(0)

                    with subprocess.Popen([res.depqbf], stdin=subprocess.PIPE, stdout=subprocess.DEVNULL, universal_newlines=True) as unsatsolver:
                        qdimacs(res.seq, unsatsolver.stdin, res.PG, blocking_clauses=[blocking], isolator=res.isolator, nautydir=res.nauty)
                        unsatsolver.stdin.close()
                        assert(unsatsolver.wait() == 20)
    else:
        qdimacs(res.seq, sys.stdout, res.PG, res.BreakID, isolator=res.isolator, nautydir=res.nauty)
