import sys
import itertools
import subprocess
import igraph
import argparse
import re

import qbfolkman

from pysat.card import CardEnc, EncType

def s2s(a, output, BreakIDbin=None, isolator=None):
    m, ar = qbfolkman.parameters(a)
    vs = [i for i in range(0, ar + m)]
    extra = 0
    ecount = qbfolkman.nchoosek(ar + m, 2)
    cliquecount = qbfolkman.nchoosek(ar + m, m)

    if BreakIDbin is not None:
        with subprocess.Popen([BreakIDbin, "-", "-no-relaxed"], stdin=subprocess.PIPE, stdout=output, universal_newlines=True) as BreakID:
            print("p cnf %d %d" % (ecount, cliquecount + extra), file=BreakID.stdin)

            for clique in itertools.combinations(vs, m):
                edges = [qbfolkman.edge(e[0], e[1], ar + m) for e in itertools.combinations(clique, 2)]

                print("%s 0" % " ".join(["-%d" % e for e in edges]), file=BreakID.stdin)
    else:
        if isolator is not None:
            isoheader = isolator.readline()
            isomatch = re.match("p cnf \d+ (\d+)", isoheader)

            assert(isomatch is not None)

            extra = int(isomatch.group(1))

        print("p cnf %d %d" % (ecount, cliquecount + extra), file=output)

        for clique in itertools.combinations(vs, m):
            edges = [qbfolkman.edge(e[0], e[1], ar + m) for e in itertools.combinations(clique, 2)]

            print("%s 0" % " ".join(["-%d" % e for e in edges]), file=output)

        if isolator is not None:
            for l in isolator:
                output.write(l)

    print("c propagator %d %d %d %s 0 0" % (len(a) * (ar + m) + ecount, 30, ecount, " ".join(["%d %d" % (i + 1, i + 1) for i in range(0, ecount)])), file=output)

    print("p cnf %d 0" % (len(a) * (ar + m) + ecount), file=output)

    for v in vs:
        eqcnf = CardEnc.equals([qbfolkman.vcolor(v, c, ar + m, ecount) for c in range(0, len(a))], encoding=EncType.pairwise)

        for c in eqcnf.clauses:
            print("%s 0" % " ".join([str(l) for l in c]), file=output)

    for c in range(0, len(a)):
        for clique in itertools.combinations(vs, a[c]):
            print("%s %s 0" % (" ".join(["-%s" % qbfolkman.edge(e[0], e[1], ar + m) for e in itertools.combinations(clique, 2)]), " ".join(["-%s" % qbfolkman.vcolor(v, c, ar + m, ecount) for v in clique])), file=output)

    print("c endpropagator", file=output)

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Generate sat-to-sat instances of unique Folkman witnesses")

    parser.add_argument("--s2sbin", "-s", help="Location of the sat-to-sat binary")
    parser.add_argument("--isolator", "-i", type=argparse.FileType('r'), help="Use the given isolator for the graph search")
    parser.add_argument("--BreakID", "-b", help="Break symmetries in the graph search using BreakID")
    parser.add_argument("seq", type=int, nargs="+", help="Input sequence")

    res = parser.parse_args()

    if res.s2sbin is not None:
        with subprocess.Popen([res.s2sbin, "-n=0"], stdin=subprocess.PIPE, stdout=subprocess.PIPE, universal_newlines=True) as s2sproc:
            m, ar = qbfolkman.parameters(res.seq)
            graphs = []
            nsolutions = 0

            s2s(res.seq, s2sproc.stdin, res.BreakID, res.isolator)
            s2sproc.stdin.close()

            for l in s2sproc.stdout:
                if l.startswith("v"):
                    values = dict([(abs(int(v)), int(v) > 0) for v in l.strip().split(" ")[1:]])
                    g = igraph.Graph(ar + m)
                    isomorphic = False

                    nsolutions += 1

                    for i in range(0, ar + m):
                        for j in range(i + 1, ar + m):
                            if values[qbfolkman.edge(i, j, ar + m)]:
                                g.add_edge(i, j)

                    for h in graphs:
                        if h.isomorphic_bliss(g):
                            isomorphic = True
                            break

                    if not isomorphic:
                        graphs.append(g)

            print("%d non-isomorphic graph(s) out of %d solutions" % (len(graphs), nsolutions))
    else:
        s2s(res.seq, sys.stdout)
